# ARC - Authenticated Remote Control
# common routines and command interpreter
#
# Author: R.Toebbicke, CERN
#
use strict;
use diagnostics;
$|=1;
our %ACL;
our $ARCLOG;
our @ARCETC;
@ARCETC=("/usr/libexec/arcd") unless @ARCETC;
our $USE_SYSLOG;
our $USERARG;
our $args;
our $proc;

chop($ENV{"SCLHOST"}=`uname -n`);
if ($ENV{"SCPRINCIPAL"}=~/^([^\.]+)\.([^\@]*)\@(.*)/) {
	$ENV{"SCUSER"}=$1;
	$ENV{"SCINSTANCE"}=$2;
	$ENV{"SCREALM"}=$3;
}
$ENV{"SCACL"}=($ENV{"SCPRINCIPAL"} eq "rcmd.$ENV{SCLHOST}\@$ENV{SCLREALM}")
		|| &aclcheck($ENV{"SCPRINCIPAL"});

# Check the logfile
# 11/10/2001 Vladimir.Bahyl@cern.ch
foreach my $etc (@ARCETC) {
  if (-r "$etc/config.arc" && -f "$etc/config.arc" && -s "$etc/config.arc") {
    require "$etc/config.arc";## no critic
    last;
  }
}

$USERARG=pop @ARGV;
foreach (@ARGV) {	# Execute remaining constant arguments
	eval $_; ## no critic
}

# Some predefined commands
$ACL{'whoami'}="ANY"; sub whoami { print $ENV{"SCPRINCIPAL"},"\n";}
$ACL{'uptime'}="ANY"; sub uptime { system "uptime"; }
$ACL{'checkv'}="AUTH"; sub checkv { system "/usr/afsws/bin/fs checkv"; }



if ( $USERARG ne '') {
	&exec_proc($USERARG);
} else {
	while(<STDIN>) {
		if (!/^\s*\#/) { &exec_proc($_);}
	}
}

1;

# &aclcheck($userid,$ACL-file);
sub aclcheck {
	if ($ENV{"SCACL"}) { return 1; }

	my ($user,$acl)=@_;
	local($_);
        my $fname;

	$acl="ACL" unless $acl;

	if ($acl eq "ANY") { return 1; }
	if ($acl eq "AUTH") { return (index($user,"\@")!=$[); }

	if ($acl eq 'ACL') {
		$fname="$ARCETC[0]/ACL";
	} elsif (substr($acl,$[,1) eq "/") {	# absolute path
		$fname=$acl;
	} else {
		foreach my $ETC (@ARCETC) {
			$fname="$ETC/$acl.acl";
			last if ( -r $fname);
		}
	}
	open(my $A, "<", $fname) || return(0);
	$user.="\n";
	while (<$A>) {
		if ($_ eq $user) {
			close($A); return 1;
		}
	}
	close($A);
	return(0);
}


sub exec_proc {
        local($_)=@_;
        my $ETC;
	local ($proc, $args);

	if (/^([a-zA-Z0-9_-]+)\s*/) {
		$proc=$1;
		$args=$';
		$proc=~s/-/_/g;
	} else { die "Unknown command: $_"; }

	if (!$ENV{"SCACL"}) {   # check for tricks in command
		&safeargs($proc);
	}

	# If procedure not (yet) defined, try require first
	if (!defined(&$proc)) {
		foreach my $etc (@ARCETC) {
			if (-r "$etc/$proc.arc") {
				$ETC=$etc;
				require "$ETC/$proc.arc";## no critic
				last;
			}
		}
	}

	if (defined(&$proc) && &aclcheck($ENV{"SCPRINCIPAL"},$ACL{$proc})) {
		&$proc($args);
	} elsif ($ENV{"SCACL"}) {
		eval $_; die "$@" if $@; ## no critic
	} else { die "unauthorized access for '$proc'";}
}

sub safeargs {
	if ("@_"=~/[;`\&\|]/) { die "Invalid arguments"; }
}

sub optargs {
	local($_)=@_;
	my ($x, $s);
	s/^\s+//;	# skip leading blanks even if no parms follow
	while($_) {
		if (/^\s*-(\w)(\S*)\s*/) {
			$x= ($2 eq '') ? 1 : $2;
			eval "\$opt_$1=\$x"; ## no critic
			$_=$';
		} else { last; }
        }
	return($_);
}

sub log {
    if (defined $USE_SYSLOG) {
        # 5/10/2001 - by Vladimir.Bahyl@cern.ch
        require Syslog;
        syslog(undef, "info|authpriv", "arc[$$]: $ENV{SCHOST} $ENV{SCPRINCIPAL} $_[0]");
    } else {
        $ARCLOG="/var/log/arc.log" unless defined($ARCLOG);
	my $now = localtime();
        my $L;
	if (! -r $ARCLOG) { open($L, ">", "$ARCLOG"); }
        my $untainted = "Found bad characters";
        if ($_[0] =~ /^([\w\d\s\/\-\&\;\:\_\@\!\#\%\^\'\"\(\)\[\]\{\}]+)/) {
            $untainted = $1;
        };
        my $schost = "BadHost";
        if ($ENV{"SCHOST"} =~ /^([\w\d\@]+)/){
            $schost = $1;
        };
        my $sprincipal = "badPrincipal";
        if ($ENV{"SCPRINCIPAL"} =~ /^([\w\d\@]+)/){
            $sprincipal = $1;
        };
	open($L,">>", "$ARCLOG") || return;
	printf($L "%s: ${schost} ${sprincipal} ${untainted}\n", $now);
	close($L);
    }
}
