#!/usr/bin/perl
# ARC - Authenticated Remote Control
# common routines and command interpreter
#
# Author: R.Toebbicke, CERN
#
use strict;
#use diagnostics;

$ENV{'PATH'} = "/bin:/usr/bin:/sbin";
delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};

our %ACL;
our $ARCLOG;
$ACL{'help'}="ANY";
our $USE_SYSLOG;
our $USERARG;
our $proc;
our $args;
our $opt_l;
our $opt_K;


sub help {
	print <<EOFHELP
Available commands:
	whoami
	uptime
	...
EOFHELP
;
}

our @ARCETC=("/usr/libexec/arcd");
$|=1;

chop($ENV{"SCLHOST"}=`uname -n`);
if ($ENV{"SCPRINCIPAL"}=~/^([^\.@\/]+)[\.\/]?([^\@]*)\@(.*)/) {
	$ENV{"SCUSER"}=$1;
	$ENV{"SCINSTANCE"}=$2;
	$ENV{"SCREALM"}=$3;
}

# Check the logfile
# 11/10/2001 Vladimir.Bahyl@cern.ch
foreach my $etc (@ARCETC) {
  if (-r "$etc/config.arc" && -f "$etc/config.arc" && -s "$etc/config.arc") {
    require "$etc/config.arc";## no critic
    last;
  }
}

$USERARG=pop @ARGV;
foreach (@ARGV) {	# Execute remaining constant arguments
    eval $_; ## no critic
}

$ENV{'PATH'}.=":/usr/afsws/bin";

# Some predefined commands
$ACL{'whoami'}="ANY"; sub whoami { print $ENV{"SCPRINCIPAL"},"\n";}
$ACL{'uptime'}="ANY"; sub uptime { system "uptime"; }
$ACL{'checkv'}="AUTH"; sub checkv { system "fs checkvolumes"; }

if ( $USERARG ne '') {
	&exec_proc($USERARG);
} else {
	while(<STDIN>) {
		if (!/^\s*\#/) { &exec_proc($_);}
	}
}

1;

# &aclcheck($userid,$ACL-file);
sub aclcheck {
    return &aclcheck_f(@_) || &aclcheck_f(@_[0]);
}

sub aclcheck_f {
        my ($user, $acl, $fname, @a);
        ($user, $acl) = @_;
	local ($_);


	$acl="ACL" unless defined($acl);
	if ($acl eq "ANY") { return 1; }
	if ($acl eq "AUTH") { return (index($user,"\@")!=$[); }

	if ($acl eq 'ACL') {
	    return 1 if $user eq "host/$ENV{SCLHOST}\@$ENV{SCLREALM}";
	    foreach my $d (@ARCETC) {
		$fname="$d/ACL";
		last if -r $fname;
	    }
	} elsif (substr($acl,$[,1) eq "/") {	# absolute path
		$fname=$acl;
	} else {
		foreach my $ETC (@ARCETC) {
			$fname="$ETC/$acl.acl";
			last if ( -r $fname);
		}
	}
	open(my $A, '<', $fname);
	@a = <$A>;
	close($A);

	return 1 if grep(/^\Q$user\E$/, @a);
	$user=~s/\@/.\@/;	# tribute to krb4...
	return 1 if grep(/^\Q$user\E$/, @a);
	return(0);
}


sub exec_proc {
        local($_)=@_;
        local($proc, $args);
        my ($ETC);
	my $super = &aclcheck($ENV{"SCPRINCIPAL"});

	if (/^([a-zA-Z0-9_-]+)\s*/) {
		$proc=$1;
		$args=$';
		$proc=~s/-/_/g;
	} else { die "Unknown command: $_"; }

	unless ($super) {   # check for tricks in command
		&safeargs($proc);
	}

	# If procedure not (yet) defined, try require first
	if (!defined(&$proc)) {
		foreach my $etc (@ARCETC) {
			if (-r "$etc/${proc}.arc") {
				$ETC=$etc;
				require "$ETC/${proc}.arc"; ## no critic
				last;
			}
		}
	}
        {
            no strict 'refs'; ## no critic
            if (defined(&$proc) && &aclcheck($ENV{"SCPRINCIPAL"},$ACL{$proc})) {
                &$proc($args);
            } elsif ($super) {
		eval $_; die "$@" if $@; ## no critic
            } else {
                &log("unauthorized access for '$proc' by $ENV{'SCPRINCIPAL'} on $ENV{'SCHOST'}");
                die "unauthorized access for '$proc'";
            }
            use strict 'refs';
        }
}

sub safeargs {
	if ("@_"=~/[;`\&\|]/) { die "Invalid arguments"; }
}

sub optargs {
        local($_)=@_;
        my ($x, $s);
	s/^\s+//;	# skip leading blanks even if no parms follow
	while($_) {
		if (/^\s*-(\w)(\S*)\s*/) {
			$x= ($2 eq '') ? 1 : $2;
			eval "\$opt_$1=\$x"; ## no critic
			$_=$';
		} else { last; }
        }
	return($_);
}

sub log {
    if (defined $USE_SYSLOG) {
        # 5/10/2001 - by Vladimir.Bahyl@cern.ch
        require Sys::Syslog;
        Sys::Syslog::syslog(undef, "info|authpriv", "arc[$$]: $ENV{SCHOST} $ENV{SCPRINCIPAL} $_[0]");
    } else {
	$ARCLOG="/var/log/arc.log" unless defined($ARCLOG);
	my $now = localtime();
        my $L;
	if (! -r $ARCLOG) { open($L,">","$ARCLOG"); }
        my $untainted = "Found bad characters";
        if ($_[0] =~ /^([\w\d\.\s\/\-\&\;\:\_\@\!\#\%\^\'\"\(\)\[\]\{\}]+)/) {
            $untainted = $1;
        };
        my $schost = "BadHost";
        if ($ENV{"SCHOST"} =~ /^([\w\d\@\/]+)/){
            $schost = $1;
        };
        my $sprincipal = "badPrincipal";
        if ($ENV{"SCPRINCIPAL"} =~ /^([\w\d\@\/]+)/){
            $sprincipal = $1;
        };
	open($L,">>", "$ARCLOG") || return;
	printf($L "%s: ${schost} ${sprincipal} ${untainted}\n", $now);
	close($L);
  }
}
